/* makeHybrid.c */
/*
genHybridFP():
Eit fingeravtrykk er satt saman av mange innerkodar. Nr ein no skal
finne eit hybridFP ser ein p kvar del(Innerkode) separat. Finn frst
feasible settet for denne innerkoden, s blir ein del av
hybridFP generert. Etter at delen av hybridFP er generert kan ein kyra
sporingsfunksjonen alg1() p den. Sj fila traceHybrid.c for forklaring
p alg1() og trace(). Etter at ein har laga alle delene av hybridFP,
byrjer ein p nytt med neste koalisjon. Kor mange koalisjoner ein har
blir gitt av konstanten NROFPIRCOL.

genCollusion():
Ein genererar og lagrar ein koalisjon av t brukarar. Sjekkar at alle
brukarane i ein koalisjon er unike. Alle piratane for kvar koalisjon
blir lagra i 2D tabellen 'pirates'. Dette er ndvendig for sporinga.

getNumbersOC();
Sj forklaring i 'oc.c'

getPartOfCC():
Tek som input ein gitt pirat sin ytterkode, samt kva for tall i ytterkoden
ein skal sj p. Ein hentar ut det aktuelle tallet fra ytterkoden, og brukar
dette for  bestemme korleis innerkoden skal sj ut. Returnerer s innerkoden.

findFeasible():
Input er den aktuelle innerkoden til alle brukarane i ein koalisjon.
Ein samanlignar s desse innerkodane, og finn det feasible settet, som
blir returnert.

makeHybrid():
Genererar eit gitt antall 64 bit blokkar tilfeldig. Antallet blokker er likt
antallet blokker i innerkoden. S sjekkar ein at hybridFP er gyldig. Ein snur
dei posisjonene som er feil. (0->1 eller 1->0)
*/

#include "bonehShaw.h"

/* Make hybrid fingerprint */
int genHybridFP(cParam p, int flagP, int flagT)
{
  int i,j,k,l,m,error;
  dword *feasible, *hybridFP, **pirateCol, *arrayptr;
  word **pirates, *arrayptr2;
  long **everyPos,*arrayptr1;
  word **pirOC, *arrayptr3;

  error=0;

  EOM((feasible = malloc(p.nrOfBlocksIC * sizeof(dword))));
  EOM((hybridFP = malloc(p.nrOfBlocksIC * sizeof(dword))));

  /* Make pirate codebook -> Allocate memory for the array */
  EOM((arrayptr = malloc(p.t * p.nrOfBlocksIC  * sizeof(dword))));
  /* Allocate room for the pointers to the rows */
  EOM((pirateCol = malloc(p.t * sizeof(dword *))));
  /* Make array to store pirate outer codes */
  EOM((arrayptr3 = malloc(p.t * p.n2 * sizeof(word))));
  EOM((pirOC = malloc(p.t * sizeof(word *))));
  /* and now we 'point' the pointers */
  for (i=0;i<p.t;i++) {
    pirateCol[i] = arrayptr + (i * p.nrOfBlocksIC);
    pirOC[i] = arrayptr3 + (i * p.n2);
  }

  /* Store the outer code numbers returned from alg1...for all the pirColls ->
     Allocate memory for the array */
  EOM((arrayptr1 = malloc(NPC * p.n2  * sizeof(long))));
  EOM((everyPos = malloc(NPC * sizeof(long))));
  /* Make pirate array */
  EOM((arrayptr2 = malloc(NPC * p.t  * sizeof(word))));
  EOM((pirates = malloc(NPC * sizeof(word))));
  for (i=0;i<NPC;i++) {
    pirates[i] = arrayptr2 + (i * (p.t));
    everyPos[i] = arrayptr1 + (i * (p.n2));
  }

  /* Create a nr of collusions -> make part of hybridfp -> trace part */
  for(i=0;i<NPC;i++) {
    genCollusion(pirates[i],p);

    for(j=0;j<p.t;j++) /* Get pirate outer code */
      getOC(pirates[i][j],pirOC[j],p);

    /* For all digits in the outer code */
    for(j=0;j<p.n2;j++) {
      /* Get one of the inner codeblocks for all pirates */
      for(k=0;k<p.t;k++) {
	getPartOfCC(pirateCol[k], pirOC[k],p, k, j);

      	if(flagP==1)	{
          printf("\nUser %d IC corresponds to nr in position %d in OC "
		 ,pirates[i][k], j);
          printBinaryCode(pirateCol[k], p);
      	}
      }

      findFeasible(feasible, pirateCol, p, flagP);
      makeHybrid(hybridFP,pirateCol,feasible,p, flagP);

      if(flagT==1)
	/* Find and return the outer code nr from alg1. (Part of tracing alg)*/
	everyPos[i][j] = alg1(hybridFP,p, flagP);
    }
  }

  if(flagT==1)
    error = trace(everyPos, pirates, p, flagP);

  free(everyPos);
  free(arrayptr);
  free(arrayptr1);
  free(pirateCol);
  free(feasible);
  free(hybridFP);
  free(pirates);

  return error;
}

/* Generate a random pirate collusion of size t */
void genCollusion(word *pirates, cParam p)
{
  int i,j;
  word randnum;

  for(i=0;i<p.t;i++) {
    pirates[i] = getRandomWord(p.M); /* Select pirate in the range: [0..M-1] */
    /* Test to ensure that all pirates in the collusion are different  */
    for(j=0;j<i;j++)
      if(pirates[j] == pirates[i])
	i=j=i-1;
  }
}

/* Returns one of the inner codewords that
make up the concatenated code for a user  */
void getPartOfCC(dword *innerCode, word *outerCode,
				cParam p, word userNr, int OCpos)
{
  int i, nrOfNull,changeBlocks, lastBlock;

  /* Set innerCode according to the number in OC */
  nrOfNull = p.r * outerCode[OCpos];

  /* How many blocks must be changed  */
  changeBlocks = (nrOfNull % BS_DWORD) == 0 ?
	       nrOfNull/BS_DWORD : (nrOfNull/BS_DWORD) + 1;

  for(i=changeBlocks;i<p.nrOfBlocksIC;i++)
    innerCode[i] = ULLONG_MAX;

  lastBlock =  nrOfNull - (BS_DWORD * (changeBlocks -1));

  for(i=0;i<changeBlocks;i++) {
    if (i==changeBlocks-1)
      innerCode[i] = ULLONG_MAX^(ULLONG_MAX<<(BS_DWORD-lastBlock));
    else
      innerCode[i] = (dword)0;
  }
}

/* Find feasible set */
void findFeasible(dword *feasible, dword **pirateCol, cParam p, int flagP)
{
  int i, j, k;

  if(flagP==1)printf("\nFinding feasible set for this inner code...");
  for(i=0;i<p.t;i++)  /* for every codeword...(part of concat code)*/
  /* and for every part of that inner codeword */
    for(j=0;j<p.nrOfBlocksIC;j++)
      for(k=i+1;k<p.t;k++) { /*...compare against the rest */
	if (k == i+1 && i == 0) /* No value in feasible[j] */
	  feasible[j] = pirateCol[i][j] ^ pirateCol[k][j];
	else
	  feasible[j] |= pirateCol[i][j] ^ pirateCol[k][j];
      }

  if(flagP==1) {
    printf("Feasible set found!\n");
    printf("Feasible: ");
    printBinaryCode(feasible, p);
    printf("\nCreating part of hybrid fingerprint...");
  }
}

/* Make part of hybrid fingerprint */
void makeHybrid(dword *hybridFP,dword **pirateCol,
				dword *feasible, cParam p, int flagP)
{
  int i;

  for(i=0;i<p.nrOfBlocksIC;i++) { /* For all blocks */
    /* Generate random block(Part of the hybrid fingerprint) */
    hybridFP[i] = getRandomDWord();

    /* Check if hybrid fingerprint is valid -> If not, correct it */
    hybridFP[i] =
    	((pirateCol[0][i] ^ hybridFP[i]) & (~feasible[i])) ^ hybridFP[i];
  }

  if(flagP==1) {
    printf("Part of hybrid fingerprint created!\n");
    printf("HybridFP: ");
    printBinaryCode(hybridFP, p);
  }
}

