/* traceHybrid.c */
/*
trace():
Nr ein har funne 'ytterkoden' til hybridFP,(Ved hjelp av alg1.)
for alle koalisjonane, sender ein desse kodane(Lagra i everyPos[][].)
med til denne funksjonen. Her blir tracePirate() kyrt. Dette er ein
funksjon som samanlignar 'ytterkoden' funne fr hybridFP til kvar
koalisjon med dei faktiske ytterkodane til alle brukarane. Den brukaren,
for kvar koalisjon, som har ytterkode likast, blir returnert.
Desse brukarane blir s sjekke opp mot dei faktiske piratane i dei ulike
koalisjonane. Om den returnerte brukaren ikkje finns i den aktuelle
koalisjonen, da blir ein feil-variabel oppdatert.

alg1():
Dette er alg1 fra Boneh&Shaw. Denne funksjonen blir kyrt fr genHybridFP()
i makeHybridFP.c. Grunnen til dette er at ein byrjar sporinga av kvar
del av hybridFP for kvar koalisjon med ein gong delen er generert. Treng da
ikkje  lagre det fullstendige hybridFP for kvar koalisjon p disk.
alg1() tek som input ein del av hybridFP og sjekkar kva for innerkode som mest
sannsynleg vart brukt for  generere denne delen. Returnerar nr p denne
innerkoden. (Som samsvarar med eit nummer i ytterkoden.)

weight():
Blir brukt av alg1() for  kalkulere vekta av eit spesielt omrde p
ein del av hybridFP. Alts antallet 1'ere over eit omrde blir summert.

tracePirat():
-Sj trace()
*/

#include "bonehShaw.h"

/* Find 'guilty' users, and check if they really are guilty */
int trace(long **everyPos, word **pirates, cParam p, int flagP)
{
  int i,j, error, pirFound, *guiltyUsers;

  error=pirFound=0;
  EOM((guiltyUsers = malloc(NPC * sizeof(int))));

  /* Find guilty users for all collusions by using CND */
  closestNDec(everyPos,guiltyUsers,p, flagP);

  /* Check if the guilty users returned really are guilty */
  for(i=0;i<NPC;i++) {
    for(j=0;j<p.t;j++)
      if (guiltyUsers[i] == pirates[i][j])
	pirFound = 1;

    if(pirFound==0){
      error++;/*printf("\n\nerror\n\n");*/}
    pirFound=0;
  }

  free(guiltyUsers);
  return error;
}


/* This is the decoding algorithm for the inner code in the B&S paper(alg1).
   We apply alg1 to each of the n2 components of the HybridFP  */
long alg1(dword *hybridFPblock, cParam p, int flagP)
{
  int s,w,k,flag;
  long pos;

  if(flagP==1) printf("ICepsilon:%0.15f\n",p.ICepsilon);
  flag = 1;

  /* Check condition 1 of alg 1. Check d first positions */
  w = weight(hybridFPblock, p, 1);
  if(flagP==1) printf("first cond. weight = %u\n", w);
  if (w > 0)
    pos = 0;
  else {
    /* Check condition 2 of alg 1. Check d last positions */
    w = weight(hybridFPblock, p, p.q-1);
    if(flagP==1) printf("second cond. weight = %u\n", w);
    if (w < p.r)
      pos = p.q-1;
    else {
      /* Check condition 3 of alg 1. */
      for(s=2; s<p.q;s++) {
	k = weight(hybridFPblock, p, s);    //B_s
	w = weight(hybridFPblock, p, s-1);  //B_(s-1)
	k = k + w;			    //B_s U B_(s-1)

	if(flagP==1) printf("third cond. weigth. k = %u\tw = %d\n", k, w);
	if (w < (double)k/2 - sqrt(((double)k/2 * p.expr))) {
	  pos = s-1;
	  flag = 0;
	  s = p.q;
	}
      }
     }
  }
  return pos;
}

/* Calculate the weight of a particular pattern(r length) in hybridFP */
word weight(dword *hybridBlock, cParam p, int s)
{
  int i,r, pos, startBlock, nrOfNrBlock, check, length, count;
  word w;
  dword mask;

  w = 0;
  count = mask = 1;
  r = p.r;
  pos = p.r * s;

  while(r>0) {
    startBlock = (pos % BS_DWORD) == 0 ?
	         (pos / BS_DWORD) - 1 : pos / BS_DWORD;
    
    nrOfNrBlock = pos - (BS_DWORD * startBlock);

    /* Set number of positions to check */
    check = (r >= nrOfNrBlock) ? nrOfNrBlock : r;

    /* 	If count equals 2, then we must read block from
	rightmost position. If count equals 1
	(first block we look at), then we must set the
	correct position to read from*/
    if(count == 1) {
      count++;
      length=(r<nrOfNrBlock)?BS_DWORD - nrOfNrBlock:BS_DWORD - check;
    }
    else
      length = 0;

    for(i=0;i<check;i++)
      if ((hybridBlock[startBlock]  &   (mask << length + i)) != 0)
	w++;

    pos = pos - check;
    r = r - check;
  }

  return w;
}

/* Find guilty users for all collusions, by using closest neighbour decoding */
void closestNDec(long **everyPos, int *guiltyUsers, cParam p, int flagP)
{
  int i,j,k,counter;
  int *prev;
  word *outercode;

  EOM((outercode = malloc(p.n2 * sizeof(word))));
  EOM((prev = malloc(NPC * sizeof(int))));

  if(flagP==1) {
    printf("\n\nOuter codeword found: ");
    for(i=0;i<NPC;i++){
      guiltyUsers[i] = -1;
      prev[i] = 0;

      for(j=0;j<p.n2;j++)
	printf("%d", everyPos[i][j]);
    }
    printf("\n\nCompare it with user codewords...\n");
  }
  else
    for(i=0;i<NPC;i++){
      guiltyUsers[i] = -1;
      prev[i] = 0;
    }

  counter = 0;

  for(i=0;i<p.M;i++) {  /* For all users */
    /* Get digits from the outerCode (For the current user) */
    getOC(i, outercode, p);

    /* Check how many positions that match */
    for(j=0;j<NPC;j++) {  /* For all collusions */
      for(k=0; k<p.n2; k++)   /* For all positions in outer code */
	if (everyPos[j][k] == outercode[k])
	  counter++;

      /* If this user codeword has more positions that match, store this user */
      if (counter > prev[j]) {
	guiltyUsers[j] = i;
	prev[j] = counter;
      }
      counter = 0;
    }
  }

  free(outercode);
  free(prev);
}





